<?php
/**
 * Class Base32
 *
 * @filesource   Base32.php
 * @author       Shannon Wynter {@link http://fremnet.net/contact}
 * @author       Smiley <smiley@chillerlan.net>
 * @copyright    Copyright (c) 2006 Shannon Wynter
 */

namespace chillerlan\Base32;

/**
 * Class to provide base32 encoding/decoding of strings
 *
 * ChangeLog
 * -----------
 * version 1.1.0, 2016-02-29, smiley {@link https://github.com/codemasher}
 *  - cleanup
 *  - separated character set class
 * version 1.0.0, 2015-12-06, smiley {@link https://github.com/codemasher}
 *  - static all the things!
 * version 0.2, 2008-08-07, Shannon Wynter {@link http://fremnet.net/contact}
 *  - Fixed transposition of Y and Z in csRFC3548
 * version 0.1, 2006-06-22, Shannon Wynter {@link http://fremnet.net/contact}
 *  - Initial release
 *
 * Notes
 * -----------
 * For dealing with humans it's probably best to use csSafe rather then csRFC3548
 */
class Base32{

	/**
	 * charset
	 *
	 * Internal holder of the current character set.
	 *
	 * @var string
	 */
	public static $charset = Base32Characters::RFC3548;

	/**
	 * setCharset
	 *
	 * Used to set the internal _charset variable
	 * I've left it so that people can arbirtrarily set their
	 * own charset
	 *
	 * @param string $charset The character set you want to use
	 *
	 * @throws \chillerlan\Base32\Base32Exception
	 */
	public static function setCharset($charset = Base32Characters::RFC3548){
		if(strlen($charset) === 32){
			self::$charset = strtoupper($charset);
		}
		else{
			throw new Base32Exception('Length must be exactly 32');
		}
	}

	/**
	 * str2bin
	 *
	 * Converts any ascii string to a binary string
	 *
	 * @param string $str The string you want to convert
	 *
	 * @return string String of 0's and 1's
	 */
	public static function str2bin($str){
		$chrs = unpack('C*', $str);

		return vsprintf(str_repeat('%08b', count($chrs)), $chrs);
	}

	/**
	 * bin2str
	 *
	 * Converts a binary string to an ascii string
	 *
	 * @param string $str The string of 0's and 1's you want to convert
	 *
	 * @return string The ascii output
	 * @throws \chillerlan\Base32\Base32Exception
	 */
	public static function bin2str($str){
		self::checkLength($str);
		self::checkBin($str);

		preg_match_all('/.{8}/', $str, $chrs);
		$chrs = array_map('bindec', $chrs[0]);

		// I'm just being slack here
		array_unshift($chrs, 'C*');

		return call_user_func_array('pack', $chrs);
	}

	/**
	 * fromBin
	 *
	 * Converts a correct binary string to base32
	 *
	 * @param string $str The string of 0's and 1's you want to convert
	 *
	 * @return string String encoded as base32
	 * @throws \chillerlan\Base32\Base32Exception
	 */
	public static function fromBin($str){
		self::checkLength($str);
		self::checkBin($str);

		// Base32 works on the first 5 bits of a byte, so we insert blanks to pad it out
		$str = preg_replace('/(.{5})/', '000$1', $str);

		// We need a string divisible by 5
		$length = strlen($str);
		$rbits = $length&7;

		if($rbits > 0){
			// Excessive bits need to be padded
			$ebits = substr($str, $length-$rbits);
			$str = substr($str, 0, $length-$rbits).'000'.$ebits.str_repeat('0', 5-strlen($ebits));
		}

		preg_match_all('/.{8}/', $str, $chrs);

		$chrs = array_map(function ($str){
			return self::$charset[bindec($str)];
		}, $chrs[0]);

		return implode('', $chrs);
	}

	/**
	 * toBin
	 *
	 * Accepts a base32 string and returns an ascii binary string
	 *
	 * @param string $str The base32 string to convert
	 *
	 * @return string Ascii binary string
	 * @throws \chillerlan\Base32\Base32Exception
	 */
	public static function toBin($str){
		self::checkCharacterSet($str);

		// Convert the base32 string back to a binary string
		$str = array_map(function ($chr){
			return sprintf('%08b', strpos(self::$charset, $chr));
		}, str_split($str));

		// Remove the extra 0's we added
		$str = preg_replace('/000(.{5})/', '$1', implode('', $str));

		// Unpad if nessicary
		$length = strlen($str);
		$rbits = $length&7;

		if($rbits > 0){
			$str = substr($str, 0, $length-$rbits);
		}

		return $str;
	}

	/**
	 * fromString
	 *
	 * Convert any string to a base32 string
	 * This should be binary safe...
	 *
	 * @param string $str The string to convert
	 *
	 * @return string The converted base32 string
	 */
	public static function fromString($str){
		return self::fromBin(self::str2bin($str));
	}

	/**
	 * toString
	 *
	 * Convert any base32 string to a normal sctring
	 * This should be binary safe...
	 *
	 * @param string $str The base32 string to convert
	 *
	 * @return string The normal string
	 */
	public static function toString($str){
		$str = strtoupper($str);

		// csSave actually has to be able to consider extra characters
		if(self::$charset === Base32Characters::CROCKFORD){
			$str = str_replace('O', '0', $str);
			$str = str_replace(['I', 'L'], '1', $str);
		}

		return self::bin2str(self::toBin($str));
	}

	/**
	 * @param string $str
	 *
	 * @throws \chillerlan\Base32\Base32Exception
	 */
	protected static function checkLength($str){
		if(strlen($str)%8 > 0){
			throw new Base32Exception('Length must be divisible by 8');
		}
	}

	/**
	 * @param string $str
	 *
	 * @throws \chillerlan\Base32\Base32Exception
	 */
	protected static function checkBin($str){
		if(!preg_match('/^[01]+$/', $str)){
			throw new Base32Exception('Only 0 and 1 are permitted');
		}
	}

	/**
	 * @param string $str
	 *
	 * @throws \chillerlan\Base32\Base32Exception
	 */
	protected static function checkCharacterSet($str){
		if(!preg_match('/^['.self::$charset.']+$/', $str)){
			throw new Base32Exception('Must match character set');
		}
	}

}
